package gpt
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Module Gpt.PartitionSource
type t = private {type_guid : Uuidm.t;partition_guid : Uuidm.t;starting_lba : int64;ending_lba : int64;attributes : int64;name : string;
}private field types for the Partition record. Can be created using the smart constructors make
val make :
?name:string ->
type_guid:Uuidm.t ->
attributes:int64 ->
int64 ->
int64 ->
(t, string) resultmake ?name ~type_guid ~attributes starting_lba ending_lba constructs a Partition.t. name is assumed to be a zero-padded utf-16le encoded string.
is_zero_partition partition is true if partition is the all-zero partition (i.e. the unused partition entry)
unmarshal buf is the partition entry encoded in the beginning of buf.